"""REST API for CADA using FastAPI."""

from contextlib import asynccontextmanager
import os
import typing

from dotenv import load_dotenv
from fastapi import FastAPI, Query
from logzero import logger
from pydantic import BaseModel
from starlette.responses import Response

import cada_prio
from cada_prio import predict
from cada_prio._version import __version__

# Load environment
env = os.environ
load_dotenv()

#: Debug mode
DEBUG = env.get("CADA_DEBUG", "false").lower() in ("true", "1")
#: Path to data / model
PATH_DATA = env.get("CADA_PATH_DATA", "/data/cada")
#: Optional path to legacy model from CADA
PATH_LEGACY = env.get("CADA_PATH_LEGACY", None)

#: The CADA models, to be loaded on startup.
GLOBAL_STATIC = {}

#: V1 API prefix.
API_V1_STR = "/api/v1"


@asynccontextmanager
async def lifespan(app: FastAPI):
    # Load the models
    all_to_hgnc, hgnc_info_by_id = predict.load_hgnc_info(
        os.path.join(PATH_DATA, "hgnc_info.jsonl")
    )
    GLOBAL_STATIC["all_to_hgnc"] = all_to_hgnc
    GLOBAL_STATIC["hgnc_info_by_id"] = hgnc_info_by_id
    graph, model = predict.load_graph_model(PATH_DATA, PATH_LEGACY)
    GLOBAL_STATIC["graph"] = graph
    GLOBAL_STATIC["model"] = model

    logger.info("see /api/v1/docs for docs")
    logger.info(
        "try: /api/v1/predict?hpo_terms=HP:0000098,HP:0000218,HP:0000486&genes=FBN1,TTN,BRCA1"
    )

    yield

    GLOBAL_STATIC.clear()


app = FastAPI(
    title="cada-prio",
    description="A phenotype-based gene prioritization tool.",
    version=__version__,
    lifespan=lifespan,
    openapi_url=f"{API_V1_STR}/openapi.json",
    docs_url=f"{API_V1_STR}/docs",
)


# Register endpoint for returning CADA version.
@app.get("/api/v1/version")
async def version():
    return Response(content=cada_prio.__version__)


class PredictionResult(BaseModel):
    rank: int
    score: float
    gene_symbol: str
    ncbi_gene_id: str
    hgnc_id: str


# Register endpoint for the prediction
@app.get("/api/v1/predict")
async def handle_predict(
    hpo_terms: typing.Annotated[str, Query(example="HP:0000098,HP:0000218,HP:0000486")],
    genes: typing.Annotated[typing.Optional[str], Query(example="FBN1,TTN,BRCA1")] = None,
):
    """Predict genes for a given set of HPO terms and optionally genes."""
    hpo_terms_list = hpo_terms.split(",")
    genes_list = genes.split(",") if genes else None
    _, sorted_scores = predict.run_prediction(
        hpo_terms_list,
        genes_list,
        GLOBAL_STATIC["all_to_hgnc"],
        GLOBAL_STATIC["graph"],
        GLOBAL_STATIC["model"],
        PATH_LEGACY is not None,
    )
    hgnc_info_by_id = GLOBAL_STATIC["hgnc_info_by_id"]
    return [
        PredictionResult(
            rank=rank,
            score=score,
            gene_symbol=hgnc_info_by_id[hgnc_id].symbol,
            ncbi_gene_id=hgnc_info_by_id[hgnc_id].ncbi_gene_id,
            hgnc_id=hgnc_info_by_id[hgnc_id].hgnc_id,
        )
        for rank, (hgnc_id, score) in enumerate(sorted_scores.items(), start=1)
    ]



"""Helpers for inspection models"""

import pickle

from logzero import logger
import networkx as nx

from cada_prio.predict import load_hgnc_info


def dump_graph(path_graph: str, path_hgnc_info: str, hgnc_to_entrez: bool):
    _, hgnc_info_by_id = load_hgnc_info(path_hgnc_info)
    hgnc_info_by_ncbi_gene_id = {
        "Entrez:%s" % x.ncbi_gene_id: x for x in hgnc_info_by_id.values() if x.ncbi_gene_id
    }
    with open(path_graph, "rb") as inputf:
        graph: nx.Graph = pickle.load(inputf)
    for edge in sorted(graph.edges):
        lhs, rhs = edge

        if hgnc_to_entrez:
            if lhs.startswith("HGNC:"):
                lhs = "Entrez:%s" % hgnc_info_by_id[lhs].ncbi_gene_id
            if rhs.startswith("HGNC:"):
                rhs = "Entrez:%s" % hgnc_info_by_id[rhs].ncbi_gene_id
        else:  # entrez to hgnc
            if lhs.startswith("Entrez:"):
                if lhs not in hgnc_info_by_ncbi_gene_id:
                    logger.warning("no HGNC ID for Entrez ID %s", lhs)
                    continue
                lhs = hgnc_info_by_ncbi_gene_id[lhs].hgnc_id
            if rhs.startswith("Entrez:"):
                if rhs not in hgnc_info_by_ncbi_gene_id:
                    logger.warning("no HGNC ID for Entrez ID %s", rhs)
                    continue
                rhs = hgnc_info_by_ncbi_gene_id[rhs].hgnc_id

        print(f"{lhs}\t{rhs}")



"""Code for parameter optimization."""

from datetime import timedelta
import gzip
import hashlib
import json
import os
import random
import time
import typing

import cattrs
from logzero import logger
import tqdm

from cada_prio import predict, train_model

#: Default seeed value.
DEFAULT_SEED = 1


def load_clinvar_phenotype_links_jsonl(path) -> typing.Iterable[train_model.GenePhenotypeRecord]:
    if path.endswith(".gz"):
        inputf = gzip.open(path, "rt")
    else:
        inputf = open(path, "rt")
    with inputf:
        for line in inputf:
            yield cattrs.structure(json.loads(line), train_model.GenePhenotypeRecord)


def make_links_unique_by_submitter(
    links: typing.Iterable[train_model.GenePhenotypeRecord],
) -> typing.List[train_model.GenePhenotypeRecord]:
    links_dict = {
        f"{link.submitter}/{','.join(link.hgnc_ids)}/{','.join(link.hpo_terms)}": link
        for link in links
        if len(link.hgnc_ids) == 1
    }
    return list(links_dict.values())


def load_links(
    path_clinvar_phenotype_links: str,
    fraction_links: typing.Optional[float],
    path_validation_links: typing.Optional[str],
    seed: typing.Optional[int],
):
    rng = random.Random(seed or DEFAULT_SEED)

    logger.info("Loading phenotype links...")
    logger.info("- load JSONL")
    orig_phenotype_links = list(load_clinvar_phenotype_links_jsonl(path_clinvar_phenotype_links))
    logger.info("- make unique by submitter")
    phenotype_links = make_links_unique_by_submitter(orig_phenotype_links)
    logger.info("- randomizing")
    rng.shuffle(phenotype_links)
    logger.info("... done loading %s links", len(phenotype_links))

    if fraction_links is None:
        assert (
            path_validation_links is not None
        ), "must give fraction_links or path_validation_links"
        links_training = phenotype_links
        orig_links_validation = list(load_clinvar_phenotype_links_jsonl(path_validation_links))
        links_validation = make_links_unique_by_submitter(orig_links_validation)
        links_test: typing.List[train_model.GenePhenotypeRecord] = []
        logger.info("Counts in explicit validation set...")
        logger.info("- training:   % 6d", len(links_training))
        logger.info(
            "- validation: % 6d (non-unique: %d)", len(links_validation), len(orig_links_validation)
        )
        logger.info("- test:       % 6d", len(links_test))
        logger.info("... that's all")
    else:
        logger.info("Computing counts...")
        n_links_total = len(phenotype_links)
        n_links_training_all = int(0.6 * n_links_total)
        n_links_training = int(n_links_training_all * fraction_links)
        n_links_validation = (n_links_total - n_links_training_all) // 2
        n_links_test = n_links_validation
        links_training = phenotype_links[:n_links_training]
        links_validation = phenotype_links[n_links_training : n_links_training + n_links_validation]
        links_test = phenotype_links[
            n_links_training
            + n_links_validation : n_links_training
            + n_links_validation
            + n_links_test
        ]
        logger.info("- total:        % 6d", len(phenotype_links))
        logger.info("- training-all: % 6d", n_links_training_all)
        logger.info("- training:     % 6d", len(links_training))
        logger.info("- validation:   % 6d", len(links_validation))
        logger.info("- test:         % 6d", len(links_test))
        logger.info("... done computing counts")

    return links_training, links_validation, links_test


def prepare_training(
    path_out: str,
    links_training: typing.List[train_model.GenePhenotypeRecord],
    links_validation: typing.List[train_model.GenePhenotypeRecord],
    links_test: typing.List[train_model.GenePhenotypeRecord],
    fraction_links: typing.Optional[float],
    path_validation_links: typing.Optional[str],
    path_embedding_params: typing.Optional[str],
    seed: typing.Optional[int],
):
    if seed is None:
        seed = DEFAULT_SEED

    logger.info("Preparing training...")
    if path_embedding_params:
        logger.info("- loading embedding params from %s", path_embedding_params)
        with open(path_embedding_params, "rt") as inputf:
            embedding_params = cattrs.structure(json.load(inputf), train_model.EmbeddingParams)
    else:
        logger.info("- using default embedding params params")
        embedding_params = train_model.EmbeddingParams(seed_embedding=seed + 23, seed_fit=seed + 42)
    params_name = hashlib.md5(
        json.dumps(cattrs.unstructure(embedding_params)).encode("utf-8")
    ).hexdigest()

    if fraction_links is None:
        path_out = f"{path_out}/explicit-validation.{seed}.{params_name}"
        _ = path_validation_links
    else:
        path_out = f"{path_out}/{int(fraction_links * 100)}.{seed}.{params_name}"

    os.makedirs(path_out, exist_ok=True)
    logger.info("- output path: %s", path_out)
    logger.info(
        "- embedding params:\n%s", json.dumps(cattrs.unstructure(embedding_params), indent=2)
    )
    path_embedding_params = f"{path_out}/embedding_params.json"
    with open(path_embedding_params, "wt") as outputf:
        print(json.dumps(cattrs.unstructure(embedding_params), indent=2), file=outputf)
    path_links_training = f"{path_out}/links_training.jsonl"
    logger.info("- trailing links: %s", path_links_training)
    with open(path_links_training, "wt") as outputf:
        for link in links_training:
            print(json.dumps(cattrs.unstructure(link)), file=outputf)
    path_links_validation = f"{path_out}/links_validation.jsonl"
    logger.info("- validation links: %s", path_links_validation)
    with open(path_links_validation, "wt") as outputf:
        for link in links_validation:
            print(json.dumps(cattrs.unstructure(link)), file=outputf)
    path_links_test = f"{path_out}/links_test.jsonl"
    logger.info("- test links: %s", path_links_test)
    with open(path_links_test, "wt") as outputf:
        for link in links_test:
            print(json.dumps(cattrs.unstructure(link)), file=outputf)
    logger.info("... done preparing training")

    return (
        path_out,
        path_links_training,
        path_links_validation,
        path_links_test,
        path_embedding_params,
    )


def run_training(
    path_out: str,
    path_hgnc_json: str,
    path_links_training: str,
    path_hpo_genes_to_phenotype: str,
    path_hpo_obo: str,
    path_embedding_params: typing.Optional[str],
    cpus: int,
):
    start = time.time()
    logger.info("Running training...")
    train_model.run(
        f"{path_out}/model",
        path_hgnc_json,
        path_links_training,
        path_hpo_genes_to_phenotype,
        path_hpo_obo,
        path_embedding_params,
        cpus=cpus,
    )
    elapsed = time.time() - start
    logger.info("... done running training in %s", timedelta(seconds=elapsed))


def run_validation(
    path_out: str,
    path_links_validation: str,
) -> typing.Dict[int, float]:
    """Run validation on the links from the path.

    :return: dictionary with top N -> percentage of links in top N
    """
    start = time.time()
    logger.info("Running validation...")
    path_model = f"{path_out}/model"
    logger.info("- model: %s", path_model)
    all_to_hgnc, _ = predict.load_hgnc_info(os.path.join(path_model, "hgnc_info.jsonl"))
    graph, model = predict.load_graph_model(path_model)
    _, hpo_id_from_alt, _ = train_model.load_hpo_ontology(os.path.join(path_model, "hp.obo"))

    logger.info("... validation steps ...")
    links_total = 0
    links_top = {
        1: 0,
        5: 0,
        10: 0,
        50: 0,
        100: 0,
    }
    pb = tqdm.tqdm(list(load_clinvar_phenotype_links_jsonl(path_links_validation)), unit=" records")
    for link_validation in pb:
        if len(link_validation.hgnc_ids) != 1:
            logger.warn(
                "skipping submission %s with %d genes",
                link_validation.scv,
                len(link_validation.hgnc_ids),
            )
            continue
        else:
            hgnc_id = link_validation.hgnc_ids[0]

        try:
            hpo_terms = list(
                sorted(set(hpo_id_from_alt.get(x, x) for x in link_validation.hpo_terms))
            )
            _, sorted_scores = predict.run_prediction(hpo_terms, None, all_to_hgnc, graph, model)
        except predict.NoValidHpoTerms:
            logger.warn("no valid HPO terms in %s (skipped)", link_validation.hpo_terms)
            continue

        links_total += 1
        ranked_genes = list(sorted_scores.keys())
        if hgnc_id in ranked_genes:
            rank = ranked_genes.index(hgnc_id) + 1
        else:
            rank = len(ranked_genes)
        for no in links_top.keys():
            if rank <= no:
                links_top[no] += 1

    result = {
        no: 100.0 * count / links_total if links_total > 0 else 0.0
        for no, count in links_top.items()
    }
    logger.info("<result>")
    print(json.dumps(result, indent=2))
    logger.info("</result>")

    elapsed = time.time() - start
    logger.info("... done running validation in %s", timedelta(seconds=elapsed))
    return result


def train_and_validate(
    *,
    path_out: str,
    path_hgnc_json: str,
    path_hpo_genes_to_phenotype: str,
    path_hpo_obo: str,
    path_clinvar_phenotype_links: str,
    fraction_links: typing.Optional[float],
    path_validation_links: typing.Optional[str],
    path_embedding_params: typing.Optional[str],
    seed: typing.Optional[int],
    cpus: int,
) -> typing.Dict[int, float]:
    """Train model and run validation on the links from the path.

    :return: dictionary with top N -> percentage of links in top N
    """

    links_training, links_validation, links_test = load_links(
        path_clinvar_phenotype_links, fraction_links, path_validation_links, seed
    )

    (
        path_out,
        path_links_training,
        path_links_validation,
        path_links_test,
        path_embedding_params,
    ) = prepare_training(
        path_out,
        links_training,
        links_validation,
        links_test,
        fraction_links,
        path_validation_links,
        path_embedding_params,
        seed,
    )
    _ = path_links_test

    run_training(
        path_out,
        path_hgnc_json,
        path_links_training,
        path_hpo_genes_to_phenotype,
        path_hpo_obo,
        path_embedding_params,
        cpus=cpus,
    )

    return run_validation(
        path_out,
        path_links_validation,
    )



import csv
from datetime import timedelta
import enum
import gzip
import itertools
import json
import os
import pickle
import shutil
import time
import typing
import warnings

import attrs
import cattrs
from gensim.models import Word2Vec
from logzero import logger
import networkx as nx
import node2vec
import pronto


@attrs.frozen(auto_attribs=True)
class GeneIds:
    """Mapping between gene IDs"""

    symbol: str
    hgnc_id: str
    ncbi_gene_id: str
    ensembl_gene_id: str


@enum.unique
class ClinSig(enum.Enum):
    BENIGN = "benign"
    LIKELY_BENIGN = "likely benign"
    UNCERTAIN = "uncertain significance"
    LIKELY_PATHOGENIC = "likely pathogenic"
    PATHOGENIC = "pathogenic"


@attrs.frozen(auto_attribs=True)
class GenePhenotypeRecord:
    """Full record from JSONL file"""

    #: RCV accession
    rcv: str
    #: SCV accession
    scv: str
    #: Clinical significance
    clinsig: typing.Optional[ClinSig]
    #: Submitter
    submitter: typing.Optional[str]
    #: Gene HGNC ID
    hgnc_ids: typing.List[str]
    #: Linked OMIM terms
    omim_terms: typing.List[str]
    #: Linked MONDO terms
    mondo_terms: typing.List[str]
    #: Linked HPO terms
    hpo_terms: typing.List[str]


@attrs.frozen(auto_attribs=True, order=True)
class GeneToPhenotypeRecord:
    """Minimal link record"""

    #: Submitter
    submitter: str
    #: Gene HGNC ID
    hgnc_id: str
    #: Linked HPO terms
    hpo_terms: typing.Tuple[str, ...]


def load_gene_phenotype_records(path: str) -> typing.Iterable[GenePhenotypeRecord]:
    if path.endswith(".gz"):
        inputf = gzip.open(path, "rt")
    else:
        inputf = open(path, "rt")

    with inputf:
        for line in inputf:
            yield cattrs.structure(json.loads(line), GenePhenotypeRecord)


def load_hpo_ontology(
    path_hpo_obo,
) -> typing.Tuple[pronto.Ontology, typing.Dict[str, str], typing.Dict[str, str]]:
    logger.info("Loading HPO OBO from %s...", path_hpo_obo)
    logger.info("- parsing ...")
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=UnicodeWarning)
        hpo_ontology = pronto.Ontology(path_hpo_obo)
    logger.info("- extracting term ids ...")
    hpo_id_from_alt: typing.Dict[str, str] = {}
    hpo_id_to_name: typing.Dict[str, str] = {}
    for term in hpo_ontology.terms():
        if not term.name or not term.id:
            continue
        hpo_id_to_name[term.id] = term.name
        for alternate_id in term.alternate_ids:
            hpo_id_from_alt[alternate_id] = term.id
            hpo_id_to_name[alternate_id] = term.name
    logger.info("... done loading %s terms", len(hpo_ontology))

    return hpo_ontology, hpo_id_from_alt, hpo_id_to_name


def load_hpo_gen2phen(
    path_hpo_genes_to_phenotype: str, ncbi_to_hgnc: typing.Dict[str, str]
) -> typing.List[GeneToPhenotypeRecord]:
    logger.info("Loading HPO gene-phenotype links from %s...", path_hpo_genes_to_phenotype)
    logger.info("- parsing records ...")
    warned_about = set()
    with open(path_hpo_genes_to_phenotype, "rt") as inputf:
        reader = csv.DictReader(inputf, delimiter="\t")
        hpo_by_gene: typing.Dict[str, typing.Set[str]] = {}
        for record in reader:
            ncbi_gene_id = record["ncbi_gene_id"]
            if ncbi_gene_id not in ncbi_to_hgnc:
                if ncbi_gene_id not in warned_about:
                    logger.warning("No HGNC entry for Entrez %s found", ncbi_gene_id)
                    warned_about.add(ncbi_gene_id)
                continue
            hgnc_id = ncbi_to_hgnc[ncbi_gene_id]
            hpo_by_gene.setdefault(hgnc_id, set()).add(record["hpo_id"])

        hpo_gene_to_phenotype_set = {
            GeneToPhenotypeRecord(
                submitter="HPO",
                hgnc_id=hgnc_id,
                hpo_terms=tuple(hpo_terms),
            )
            for hgnc_id, hpo_terms in hpo_by_gene.items()
        }
    logger.info("- sorting ...")
    hpo_gene_to_phenotype = list(sorted(hpo_gene_to_phenotype_set))
    logger.info("... done loading %s records", len(hpo_gene_to_phenotype))

    return hpo_gene_to_phenotype


def load_clinvar_gen2phen(path_gene_hpo_links: str) -> typing.List[GeneToPhenotypeRecord]:
    logger.info("Loading ClinVar gene-phenotype links from %s...", path_gene_hpo_links)
    logger.info("- parsing records ...")
    clinvar_gene_to_phenotype_set = {
        GeneToPhenotypeRecord(
            submitter=record.submitter,
            hgnc_id=record.hgnc_ids[0],
            hpo_terms=tuple(record.hpo_terms),
        )
        for record in load_gene_phenotype_records(path_gene_hpo_links)
        if record.clinsig in (ClinSig.LIKELY_PATHOGENIC, ClinSig.PATHOGENIC)
        and len(record.hgnc_ids) == 1
        and record.submitter
    }
    logger.info("- sorting ...")
    clinvar_gene_to_phenotype = list(sorted(clinvar_gene_to_phenotype_set))
    logger.info("... done loading %s records", len(clinvar_gene_to_phenotype_set))

    return clinvar_gene_to_phenotype


def load_hgnc_info(path_hgnc_json) -> typing.Tuple[typing.Dict[str, str], typing.List[GeneIds]]:
    logger.info("Loading HGNC info from %s...", path_hgnc_json)
    gene_ids = []
    ncbi_to_hgnc = {}
    with open(path_hgnc_json) as f:
        full_hgnc = json.load(f)
        for doc in full_hgnc["response"]["docs"]:
            if all(key in doc for key in ("symbol", "entrez_id")):
                gene_id = GeneIds(
                    symbol=doc["symbol"],
                    hgnc_id=doc["hgnc_id"],
                    ncbi_gene_id=doc["entrez_id"],
                    ensembl_gene_id=doc.get("ensembl_gene_id"),
                )
                ncbi_to_hgnc[gene_id.ncbi_gene_id] = gene_id.hgnc_id
                gene_ids.append(gene_id)
    logger.info("... done loading %s records", len(gene_ids))

    return ncbi_to_hgnc, gene_ids


#: type alias for an edge.
Edge = typing.Tuple[str, str]


def yield_hpo_edges(hpo_ontology: pronto.Ontology) -> typing.Iterator[Edge]:
    """Yield hierarchical edges from the HPO ontology"""
    for term in hpo_ontology.terms():
        parents = list(term.superclasses(distance=1))
        for parent in parents[1:]:
            yield (term.id, parent.id)


def yield_gene2phen_edges(links: typing.List[GeneToPhenotypeRecord]) -> typing.Iterator[Edge]:
    """Yield edges from link record list"""
    for link in links:
        for hpo_term in link.hpo_terms:
            yield (link.hgnc_id, hpo_term)


@attrs.frozen(auto_attribs=True)
class EmbeddingParams:
    """Parameters for the embedding"""

    #: The number of dimensions of feature vectors
    dimensions: int = 300
    #: The number of nodes in each random walk
    walk_length: int = 60
    #: Controls the probability for a walk to visit immediately back to the previous node
    p: float = 1.7987535798694703
    #: Controls the probability for a walk to visit previously unexplored neighborhoods in the
    #: graph
    q: float = 3.875406134463754
    #: Number of random walks to be generated from each node in the graph
    num_walks: int = 10
    #: Limit on the number of words in each context
    window: int = 4
    #: Set the min_count in the fitting
    min_count: int = 1
    #: Set the batch_words in the fitting
    batch_words: int = 4
    #: RNG seed for embedding
    seed_embedding: int = 1
    #: Whether to use skipgram for model fitting instead of CBOW
    use_skipgram: bool = False
    #: Number of epochs for fitting
    epochs: int = 5
    #: RNG seed for fitting
    seed_fit: int = 1


def log_graph_examples(edges: typing.List[Edge]):
    edges = list(sorted(edges))
    all_nodes = set(itertools.chain([e[0] for e in edges], [e[1] for e in edges]))
    hgnc_nodes = list(sorted([v for v in all_nodes if v.startswith("HGNC:")]))
    hpo_nodes = list(sorted([v for v in all_nodes if v.startswith("HP:")]))
    logger.info("   - ten HGNC nodes: %s", hgnc_nodes[:10])
    logger.info("   - ten HPO nodes: %s", hpo_nodes[:10])


def build_and_fit_model(
    *,
    clinvar_gen2phen,
    hpo_gen2phen,
    hpo_ontology,
    embedding_params: EmbeddingParams,
    cpus: int = 1,
):
    # create graph edges combining HPO hierarchy and training edges from ClinVar
    logger.info("Constructing training graph ...")
    logger.info("- building edges ...")
    training_edges = list(
        itertools.chain(
            yield_hpo_edges(hpo_ontology),
            yield_gene2phen_edges(hpo_gen2phen),
            yield_gene2phen_edges(clinvar_gen2phen),
        )
    )
    log_graph_examples(training_edges)
    logger.info("- graph construction")
    training_graph = nx.Graph()
    training_graph.add_edges_from(training_edges)
    logger.info("... done constructing training graph with %s edges", len(training_edges))

    logger.info("Computing the embedding / model fit...")
    logger.info("- embedding graph...")
    logger.info(
        "  - using parameters:\n%s", json.dumps(cattrs.unstructure(embedding_params), indent=2)
    )
    embedding_start = time.time()
    embedding = node2vec.Node2Vec(
        training_graph,
        dimensions=embedding_params.dimensions,
        walk_length=embedding_params.walk_length,
        num_walks=embedding_params.num_walks,
        p=embedding_params.p,
        q=embedding_params.q,
        workers=cpus,
        seed=embedding_params.seed_embedding,
    )
    embedding_elapsed = time.time() - embedding_start
    logger.info("  embedding took %s", timedelta(seconds=embedding_elapsed))
    fit_start = time.time()
    logger.info("- fitting model...")
    model = embedding.fit(
        window=embedding_params.window,
        min_count=embedding_params.min_count,
        batch_words=embedding_params.batch_words,
        seed=embedding_params.seed_fit,
        sg=1 if embedding_params.use_skipgram else 0,
        epochs=embedding_params.epochs,
        workers=cpus,
    )
    fit_elapsed = time.time() - fit_start
    logger.info("  fitting took %s", timedelta(seconds=fit_elapsed))
    logger.info("... done computing the embedding")
    return training_graph, model, embedding_params


def write_graph_and_model(
    path_out,
    hgnc_info: typing.List[GeneIds],
    training_graph: nx.Graph,
    embedding_params: EmbeddingParams,
    model: Word2Vec,
):
    os.makedirs(path_out, exist_ok=True)

    path_hgnc_info = os.path.join(path_out, "hgnc_info.jsonl")
    logger.info("Saving HGNC info to %s...", path_hgnc_info)
    with open(path_hgnc_info, "w") as f:
        for record in hgnc_info:
            json.dump(cattrs.unstructure(record), f)
            f.write("\n")
    logger.info("... done saving HGNC info")

    path_graph = os.path.join(path_out, "graph.gpickle")
    logger.info("Saving graph to %s...", path_graph)
    with open(path_graph, "wb") as outputf:
        pickle.dump(training_graph, outputf)
    logger.info("... done saving graph")

    logger.info("Saving embedding to %s...", path_out)
    path_embeddings = os.path.join(path_out, "embedding")
    logger.info("- %s", path_embeddings)
    model.wv.save_word2vec_format(path_embeddings)
    path_model = os.path.join(path_out, "model")
    logger.info("- %s", path_model)
    model.save(path_model)
    path_params = os.path.join(path_out, "embedding_params.json")
    logger.info("- %s", path_params)
    with open(path_params, "wt") as outputf:
        print(json.dumps(cattrs.unstructure(embedding_params), indent=2), file=outputf)
    logger.info("... done saving embedding")


def load_embedding_params(path_embedding_params: typing.Optional[str]):
    if path_embedding_params:
        logger.info("Loading embedding parameters from %s...", path_embedding_params)
        with open(path_embedding_params, "rt") as inputf:
            embedding_params_dict = json.load(inputf)
        logger.info("... done loading embedding parameters")
    else:
        logger.info("Using default embedding parameters")
        embedding_params_dict = {}

    return cattrs.structure(embedding_params_dict, EmbeddingParams)


def run(
    path_out: str,
    path_hgnc_json: str,
    path_gene_hpo_links: str,
    path_hpo_genes_to_phenotype: str,
    path_hpo_obo: str,
    path_embedding_params: typing.Optional[str] = None,
    cpus: int = 1,
):
    # load all data
    embedding_params = load_embedding_params(path_embedding_params)
    ncbi_to_hgnc, hgnc_info = load_hgnc_info(path_hgnc_json)
    clinvar_gen2phen = load_clinvar_gen2phen(path_gene_hpo_links)
    hpo_gen2phen = load_hpo_gen2phen(path_hpo_genes_to_phenotype, ncbi_to_hgnc)
    hpo_ontology, hpo_id_from_alt, hpo_id_to_name = load_hpo_ontology(path_hpo_obo)
    _, _ = hpo_id_from_alt, hpo_id_to_name

    # build and fit model
    training_graph, model, embedding_params = build_and_fit_model(
        clinvar_gen2phen=clinvar_gen2phen,
        hpo_gen2phen=hpo_gen2phen,
        hpo_ontology=hpo_ontology,
        embedding_params=embedding_params,
        cpus=cpus,
    )
    # write out graph and model
    write_graph_and_model(path_out, hgnc_info, training_graph, embedding_params, model)

    shutil.copyfile(path_hpo_obo, f"{path_out}/hp.obo")



"""Console script for CADA"""

import json
import logging
import sys
import tempfile
import typing

import cattr
import click
import logzero
import yaml

try:
    import optuna

    _ = optuna
    HAVE_OPTUNA = True
except ImportError:
    HAVE_OPTUNA = False

from cada_prio import _version, inspection, param_opt, predict, train_model


@click.group()
@click.version_option(_version.__version__)
@click.option("--verbose/--no-verbose", default=False)
@click.pass_context
def cli(ctx: click.Context, verbose: bool):
    """Main entry point for CLI via click."""
    ctx.ensure_object(dict)
    ctx.obj["verbose"] = verbose
    if verbose:
        logzero.loglevel(logging.DEBUG)
    else:
        logzero.loglevel(logging.INFO)


@cli.command("train")
@click.argument("path_out", type=str)
@click.option("--path-hgnc-json", type=str, help="path to HGNC JSON", required=True)
@click.option(
    "--path-gene-hpo-links", type=str, help="path to gene-HPO term links JSONL", required=True
)
@click.option(
    "--path-hpo-genes-to-phenotype",
    type=str,
    help="path to genes_to_phenotype.txt file",
    required=True,
)
@click.option("--path-hpo-obo", type=str, help="path HPO OBO file", required=True)
@click.option(
    "--path-embedding-params", type=str, help="optional path to JSON file with embedding parameters"
)
@click.option("--cpus", type=int, help="number of CPUs to use", default=1)
@click.pass_context
def cli_train(
    ctx: click.Context,
    path_out: str,
    path_hgnc_json: str,
    path_gene_hpo_links: str,
    path_hpo_genes_to_phenotype: str,
    path_hpo_obo: str,
    path_embedding_params: typing.Optional[str],
    cpus: int,
):
    """train model"""
    ctx.ensure_object(dict)
    train_model.run(
        path_out,
        path_hgnc_json,
        path_gene_hpo_links,
        path_hpo_genes_to_phenotype,
        path_hpo_obo,
        path_embedding_params,
        cpus,
    )


@cli.command("predict")
@click.argument("path_model", type=str)
@click.option(
    "--hpo-terms",
    type=str,
    help="comma-separate HPO terms or @file with space-separated ones",
    required=True,
)
@click.option(
    "--genes",
    type=str,
    help="comma-separated genes to restrict prediction to or @file with space-separated ones",
)
@click.pass_context
def cli_predict(
    ctx: click.Context,
    path_model: str,
    hpo_terms: str,
    genes: typing.Optional[str],
):
    """perform prediction/prioritication"""
    ctx.ensure_object(dict)

    if hpo_terms.startswith("@"):
        with open(hpo_terms[1:]) as f:
            hpo_term_list = [x.strip() for x in f.read().split()]
    else:
        hpo_term_list = [x.strip() for x in hpo_terms.split(",")]

    gene_list = None
    if genes:
        if genes.startswith("@"):
            with open(genes[1:]) as f:
                gene_list = [x.strip() for x in f.read().split()]
        else:
            gene_list = [x.strip() for x in genes.split(",")]

    if predict.run(path_model, hpo_term_list, gene_list) != 0:
        ctx.exit(1)


@cli.group("utils")
def cli_utils():
    """utilities"""


@cli_utils.command("dump-graph")
@click.option(
    "--hgnc-to-entrez/--no-hgnc-to-entrez", help="enable HGNC to Entrez mapping", default=True
)
@click.argument("path_graph", type=str)
@click.argument("path_hgnc_info", type=str)
@click.pass_context
def cli_dump_graph(
    ctx: click.Context,
    path_graph: str,
    path_hgnc_info: str,
    hgnc_to_entrez: bool,
):
    """dump graph edges for debugging"""
    ctx.ensure_object(dict)
    inspection.dump_graph(path_graph, path_hgnc_info, hgnc_to_entrez)


@cli_utils.command("dump-openapi-yaml")
@click.argument("path_out", type=str)
def cli_dump_openapi_yaml(path_out: str):
    """Dump OpenAPI YAML file"""
    from cada_prio import rest_server

    with open(path_out, "wt") as f:
        yaml.dump(rest_server.app.openapi(), f)


@cli.group("tune")
def cli_tune():
    """hyperparameter tuning"""


@cli_tune.command("train-eval")
@click.argument("path_out", type=str)
@click.option("--path-hgnc-json", type=str, help="path to HGNC JSON", required=True)
@click.option(
    "--path-hpo-genes-to-phenotype",
    type=str,
    help="path to genes_to_phenotype.txt file",
    required=True,
)
@click.option("--path-hpo-obo", type=str, help="path HPO OBO file", required=True)
@click.option(
    "--path-clinvar-phenotype-links",
    type=str,
    help="path to ClinVar phenotype links JSONL",
    required=True,
)
@click.option(
    "--fraction-links",
    type=float,
    help="fraction of links to add to the graph (conflicts with --path-validation-links)",
)
@click.option(
    "--path-validation-links",
    type=str,
    help="path to validation links JSONL (conflicts with --fraction-links)",
)
@click.option(
    "--path-embedding-params", type=str, help="path to JSON file with embedding params; optional"
)
@click.option(
    "--seed",
    type=int,
    help="seed for random number generator",
)
@click.option("--cpus", type=int, help="number of CPUs to use", default=1)
@click.pass_context
def cli_train_eval(
    ctx: click.Context,
    path_out: str,
    path_hgnc_json: str,
    path_hpo_genes_to_phenotype: str,
    path_hpo_obo: str,
    path_clinvar_phenotype_links: str,
    fraction_links: typing.Optional[float],
    path_validation_links: typing.Optional[str],
    path_embedding_params: typing.Optional[str],
    seed: typing.Optional[int],
    cpus: int,
):
    """train and evaluate model for one set of parameters"""
    if bool(fraction_links) == bool(path_validation_links):
        raise click.UsageError(
            "exactly one of --fraction-links and --path-validation-links must be given"
        )

    ctx.ensure_object(dict)
    param_opt.train_and_validate(
        path_out=path_out,
        path_hgnc_json=path_hgnc_json,
        path_hpo_genes_to_phenotype=path_hpo_genes_to_phenotype,
        path_hpo_obo=path_hpo_obo,
        path_clinvar_phenotype_links=path_clinvar_phenotype_links,
        fraction_links=fraction_links,
        path_validation_links=path_validation_links,
        path_embedding_params=path_embedding_params,
        seed=seed,
        cpus=cpus,
    )


@cli_tune.command("eval-only")
@click.argument("path_model", type=str)
@click.argument(
    "path_validation_links",
    type=str,
)
@click.pass_context
def cli_eval_only(
    ctx: click.Context,
    path_model: str,
    path_validation_links: str,
):
    """run a trained model on validation links"""
    ctx.ensure_object(dict)
    param_opt.run_validation(
        path_out=path_model,
        path_links_validation=path_validation_links,
    )


if HAVE_OPTUNA:

    @cli_tune.command("run-optuna")
    @click.argument("storage", type=str)
    @click.option("--n-trials", type=int, help="number of trials to run; default: 100", default=100)
    @click.option(
        "--study-name",
        type=str,
        help="name of Optuna study; default: cada-tune",
        default="cada-tune",
    )
    @click.option("--path-hgnc-json", type=str, help="path to HGNC JSON", required=True)
    @click.option(
        "--path-hpo-genes-to-phenotype",
        type=str,
        help="path to genes_to_phenotype.txt file",
        required=True,
    )
    @click.option("--path-hpo-obo", type=str, help="path HPO OBO file", required=True)
    @click.option(
        "--path-clinvar-phenotype-links",
        type=str,
        help="path to ClinVar phenotype links JSONL",
        required=True,
    )
    @click.option(
        "--fraction-links",
        type=float,
        help="fraction of links to add to the graph (conflicts with --path-validation-links)",
    )
    @click.option(
        "--path-validation-links",
        type=str,
        help="path to validation links JSONL (conflicts with --fraction-links)",
    )
    @click.option(
        "--seed",
        type=int,
        help="seed for random number generator",
    )
    @click.option("--cpus", type=int, help="number of CPUs to use", default=1)
    def cli_run_optuna(
        storage: str,
        study_name: str,
        n_trials: int,
        path_hgnc_json: str,
        path_hpo_genes_to_phenotype: str,
        path_hpo_obo: str,
        path_clinvar_phenotype_links: str,
        fraction_links: typing.Optional[float],
        path_validation_links: typing.Optional[str],
        seed: typing.Optional[int],
        cpus: int,
    ):
        """run hyperparameter tuning"""

        def objective(trial: optuna.trial.BaseTrial) -> float:
            with tempfile.TemporaryDirectory() as tmpdir, open(
                f"{tmpdir}/params.json", "wt"
            ) as tmpf:
                json.dump(
                    cattr.unstructure(
                        train_model.EmbeddingParams(
                            dimensions=trial.suggest_int("dimensions", low=100, high=500, step=10),
                            walk_length=trial.suggest_int("walk_length", low=1, high=100),
                            p=trial.suggest_float("p", low=0.1, high=2.5),
                            q=trial.suggest_float("q", low=0.0, high=1.0),
                            num_walks=trial.suggest_int("num_walks", low=10, high=50),
                            window=trial.suggest_int("window", low=4, high=8),
                            epochs=trial.suggest_int("epochs", low=3, high=6),
                            use_skipgram=trial.suggest_categorical("use_skipgram", [True, False]),
                            min_count=trial.suggest_int("min_count", low=1, high=5),
                            batch_words=trial.suggest_int("batch_words", low=3, high=6),
                        )
                    ),
                    tmpf,
                    indent=2,
                )
                tmpf.flush()
                result = param_opt.train_and_validate(
                    path_out=f"{tmpdir}/out",
                    path_hgnc_json=path_hgnc_json,
                    path_hpo_genes_to_phenotype=path_hpo_genes_to_phenotype,
                    path_hpo_obo=path_hpo_obo,
                    path_clinvar_phenotype_links=path_clinvar_phenotype_links,
                    fraction_links=fraction_links,
                    path_validation_links=path_validation_links,
                    path_embedding_params=f"{tmpdir}/params.json",
                    seed=seed,
                    cpus=cpus,
                )
                return result[100]

        optuna.logging.get_logger("optuna").addHandler(logging.StreamHandler(sys.stdout))
        study = optuna.create_study(study_name=study_name, storage=storage, load_if_exists=True)
        study.optimize(objective, n_trials=n_trials)



"""Model-based prediction"""

import json
import os
import pickle
import typing

import cattrs
from gensim.models import Word2Vec
from logzero import logger
import numpy as np

from cada_prio import train_model


def load_hgnc_info(path_hgnc_jsonl):
    logger.info("Loading HGNC info...")
    logger.info("- parsing")
    hgnc_infos = []
    with open(path_hgnc_jsonl, "rt") as f:
        for line in f:
            hgnc_infos.append(cattrs.structure(json.loads(line), train_model.GeneIds))
    logger.info("- create mapping")
    all_to_hgnc = {}
    for record in hgnc_infos:
        all_to_hgnc[record.symbol] = record
        all_to_hgnc[record.ncbi_gene_id] = record
        all_to_hgnc[record.hgnc_id] = record
        if record.ensembl_gene_id:
            all_to_hgnc[record.ensembl_gene_id] = record
    hgnc_info_by_id = {record.hgnc_id: record for record in hgnc_infos}
    logger.info("... done loading HGNC info")
    return all_to_hgnc, hgnc_info_by_id


def load_graph_model(path_model: str, legacy_path: typing.Optional[str] = None):
    path_graph = os.path.join(path_model, "graph.gpickle")
    path_embedding = os.path.join(path_model, "model")
    if legacy_path:
        logger.info("(using legacy model paths from %s)", legacy_path)
        path_graph = os.path.join(
            legacy_path, "data", "processed", "knowledge_graph", "unweighted", "train100.gpickle"
        )
        path_embedding = os.path.join(legacy_path, "models", "unweighted", "node2vec.model")
    logger.info("Loading graph...")
    with open(path_graph, "rb") as inputf:
        graph = pickle.load(inputf)
    logger.info("... done loading graph")
    logger.info("Loading model...")
    model = Word2Vec.load(path_embedding)
    logger.info("... done loading model")
    return graph, model


class NoValidHpoTerms(ValueError):
    pass


def run_prediction(
    orig_hpo_terms, orig_genes, all_to_hgnc, graph, model, translate_legacy_entrez_ids: bool = False
) -> typing.Tuple[typing.List[str], typing.Dict[str, float]]:
    # Lookup HPO term embeddings.
    hpo_terms = {}
    for hpo_term in orig_hpo_terms:
        if hpo_term not in model.wv:
            logger.warn("skipping HPO term %s as it is not in the model", hpo_term)
        else:
            hpo_terms[hpo_term] = model.wv[hpo_term]
    if not hpo_terms:
        logger.error("no valid HPO terms in model")
        raise NoValidHpoTerms("no valid HPO terms in query")

    # Map gene IDs to HGNC IDs
    genes = []
    for orig_gene in orig_genes or []:
        if orig_gene in all_to_hgnc:
            genes.append(all_to_hgnc[orig_gene].hgnc_id)
        else:
            genes.append(orig_gene)

    # Generate a score for each gene in the knowledge graph
    logger.debug("Generating scores...")
    gene_scores = {}
    for node_id in graph.nodes():
        if translate_legacy_entrez_ids and node_id.startswith("Entrez:"):
            graph_gene_id = node_id
            stripped_node_id = node_id[len("Entrez:") :]
            if stripped_node_id not in all_to_hgnc:
                logger.warning(
                    "skipping legacy gene id %s as cannot translate to HGNC", graph_gene_id
                )
                continue
            hgnc_id = all_to_hgnc[stripped_node_id].hgnc_id
        elif node_id.startswith("HGNC:"):  # is gene
            graph_gene_id = node_id
            hgnc_id = node_id
        else:
            continue  # skip, no gene

        # if we reach here then we have a gene
        if genes and hgnc_id not in genes:
            continue  # skip, not asked for

        this_gene_scores = []
        graph_gene_emb = model.wv[graph_gene_id]
        for hpo_term, hpo_term_emb in hpo_terms.items():
            score = np.dot(hpo_term_emb, graph_gene_emb)
            this_gene_scores.append(score)
        gene_scores[hgnc_id] = sum(this_gene_scores) / len(hpo_terms)

    sorted_scores = dict(sorted(gene_scores.items(), key=lambda x: x[1], reverse=True))
    return list(hpo_terms.keys()), sorted_scores


def run(
    path_model: str,
    orig_hpo_terms: typing.List[str],
    orig_genes: typing.Optional[typing.List[str]] = None,
) -> int:
    # Load and prepare data
    all_to_hgnc, hgnc_info_by_id = load_hgnc_info(os.path.join(path_model, "hgnc_info.jsonl"))
    graph, model = load_graph_model(path_model)
    try:
        hpo_terms, sorted_scores = run_prediction(
            orig_hpo_terms, orig_genes, all_to_hgnc, graph, model
        )
    except NoValidHpoTerms:
        return 1

    # Write out results to stdout, largest score first
    print("# query (len=%d): %s" % (len(hpo_terms), ",".join(hpo_terms)))
    print("\t".join(["rank", "score", "gene_symbol", "ncbi_gene_id", "hgnc_id"]))
    for rank, (hgnc_id, score) in enumerate(sorted_scores.items(), start=1):
        hgnc_info = hgnc_info_by_id[hgnc_id]
        print(
            "\t".join(
                map(
                    str,
                    [
                        rank,
                        score,
                        hgnc_info.symbol,
                        hgnc_info.ncbi_gene_id,
                        hgnc_info.hgnc_id,
                    ],
                )
            )
        )

    return 0
